/**
 *	@file Main.cpp
 *	@brief OpenGL extension compiler
 *	@author -tHE SWINe-
 *
 *	@date 2011-07-15
 *
 *	Added checking for double requests for the same extension.
 *	Added simple support for number tags in token names (eg. GL_TEXTURE<n>).
 *	Added the OpenGL autocheck feature.
 *	Fixed some minor parser errors.
 *	Forced function argument names (using __unnamed_x for parameters with unkown names).
 *
 *	@date 2011-07-20
 *
 *	Improved autocheck for statically-linked void functions so if using the user breakpoint,
 *	the code displayed in MSVC is user code (the call that caused the error) and not
 *	OpenGL40.h. This only increases user comfort, functionality didn't change.
 *
 *	Added the #tab-size final assembly command.
 *
 *	Removed tabs from all the files in ./src folder.
 *
 *	Added the possibility to expand certain macros where possible to reduce compilation
 *	time for the resulting times.
 */

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <stdio.h>
#include <string.h>
#include <vector>
#include <string>
#include <set>
#include <map>
#include "../UberLame_src/Thread.h"
#include "../UberLame_src/MinMax.h"
#include "../UberLame_src/StlUtils.h"
#include "../UberLame_src/Dir.h"
#include "Download.h"
#include "GLExtension.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

// process command buffer and create the three intermediate files
static bool Process_Commands(const char **p_command_list, size_t n_command_num,
	FILE *p_header_file, FILE *p_getters_file, FILE *p_source_file,
	bool b_force_downloads, bool b_dont_notify_cached_files,
	bool b_want_gl_autocheck);

// process command buffer and assemble output files from prepared source and generated intermediates
static bool Final_Assembly(const char **p_command_list, size_t n_command_num);

int main(int n_arg_num, const char **p_arg_list)
{
	const char *p_command_list[] = {
		"#autodoc enable",
		// enable documentation for old, documented extensions

		"#copydoc Uniform1d Uniform1f",
		"#copydoc Uniform2d Uniform2f",
		"#copydoc Uniform3d Uniform3f",
		"#copydoc Uniform4d Uniform4f",
		"#copydoc Uniform1dv Uniform1fv",
		"#copydoc Uniform2dv Uniform2fv",
		"#copydoc Uniform3dv Uniform3fv",
		"#copydoc Uniform4dv Uniform4fv",
		"#copydoc ProgramUniform1d ProgramUniform1f",
		"#copydoc ProgramUniform2d ProgramUniform2f",
		"#copydoc ProgramUniform3d ProgramUniform3f",
		"#copydoc ProgramUniform4d ProgramUniform4f",
		"#copydoc ProgramUniform1dv ProgramUniform1fv",
		"#copydoc ProgramUniform2dv ProgramUniform2fv",
		"#copydoc ProgramUniform3dv ProgramUniform3fv",
		"#copydoc ProgramUniform4dv ProgramUniform4fv",
		"#copydoc UniformMatrix2dv UniformMatrix2fv",
		"#copydoc UniformMatrix3dv UniformMatrix3fv",
		"#copydoc UniformMatrix4dv UniformMatrix4fv",
		"#copydoc UniformMatrix2x3dv UniformMatrix2x3fv",
		"#copydoc UniformMatrix3x2dv UniformMatrix3x2fv",
		"#copydoc UniformMatrix2x4dv UniformMatrix2x4fv",
		"#copydoc UniformMatrix4x2dv UniformMatrix4x2fv",
		"#copydoc UniformMatrix3x4dv UniformMatrix3x4fv",
		"#copydoc UniformMatrix4x3dv UniformMatrix4x3fv",
		"#copydoc ProgramUniformMatrix2dv ProgramUniformMatrix2fv",
		"#copydoc ProgramUniformMatrix3dv ProgramUniformMatrix3fv",
		"#copydoc ProgramUniformMatrix4dv ProgramUniformMatrix4fv",
		"#copydoc ProgramUniformMatrix2x3dv ProgramUniformMatrix2x3fv",
		"#copydoc ProgramUniformMatrix3x2dv ProgramUniformMatrix3x2fv",
		"#copydoc ProgramUniformMatrix2x4dv ProgramUniformMatrix2x4fv",
		"#copydoc ProgramUniformMatrix4x2dv ProgramUniformMatrix4x2fv",
		"#copydoc ProgramUniformMatrix3x4dv ProgramUniformMatrix3x4fv",
		"#copydoc ProgramUniformMatrix4x3dv ProgramUniformMatrix4x3fv",
		"#copydoc GetUniformuiv GetUniformiv",
		"#copydoc GetUniformdv GetUniformfv",
		"#copydoc GetTexParameterIiv GetTexParameteriv",
		"#copydoc GetTexParameterIuiv GetTexParameteriv",
		"#copydoc GetBufferParameteri64v GetBufferParameteriv",
		"#copydoc SamplerParameterIiv SamplerParameteriv",
		"#copydoc SamplerParameterIuiv SamplerParameteriv",
		"#copydoc GetSamplerParameterIiv GetSamplerParameteriv",
		"#copydoc GetSamplerParameterIuiv GetSamplerParameteriv",
		"#copydoc VertexAttribP1uiv VertexAttrib1sv",
		"#copydoc VertexAttribP2uiv VertexAttrib2sv",
		"#copydoc VertexAttribP3uiv VertexAttrib3sv",
		"#copydoc VertexAttribP4uiv VertexAttrib4sv", // *v versions yes, simple P versions no
		"#copydoc EnableIndexed Enablei",
		"#copydoc DisableIndexed Disablei",
		"#copydoc IsEnabledIndexed IsEnabledi",
		"#copydoc GetFloati_v GetIntegeri_v",
		"#copydoc GetDoublei_v GetIntegeri_v",
		// undocumented functions with very similar documented functions

		"#docu-location EnableIndexed http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml",
		"#docu-location DisableIndexed http://www.opengl.org/sdk/docs/man4/xhtml/glDisable.xml",
		"#docu-location IsEnabledIndexed http://www.opengl.org/sdk/docs/man4/xhtml/glIsEnabled.xml",
		"#docu-location GetBooleanv http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetBooleani_v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetDoublev http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetFloatv http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetIntegerv http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetIntegeri_v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetFloati_v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetDoublei_v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetInteger64i_v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetIntegerIndexediv http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetInteger64v http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetString http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location UniformMatrix2fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix2x3fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3x2fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix2x4fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4x2fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3x4fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4x3fv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix2dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix2x3dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3x2dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix2x4dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4x2dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix3x4dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location UniformMatrix4x3dv http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml",
		"#docu-location ProgramUniformMatrix2fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix2x3fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3x2fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix2x4fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4x2fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3x4fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4x3fv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix2dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix2x3dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3x2dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix2x4dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4x2dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix3x4dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location ProgramUniformMatrix4x3dv http://www.opengl.org/sdk/docs/man4/xhtml/glProgramUniform.xml",
		"#docu-location VertexAttribIPointer http://www.opengl.org/sdk/docs/man4/xhtml/glVertexAttribPointer.xml",
		"#docu-location VertexAttribLPointer http://www.opengl.org/sdk/docs/man4/xhtml/glVertexAttribPointer.xml",
		"#docu-location Disable http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml",
		"#docu-location Disablei http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml",
		"#docu-location EndQuery http://www.opengl.org/sdk/docs/man4/xhtml/glBeginQuery.xml",
		"#docu-location UnmapBuffer http://www.opengl.org/sdk/docs/man4/xhtml/glMapBuffer.xml",
		"#docu-location ColorMaski http://www.opengl.org/sdk/docs/man4/xhtml/glColorMask.xml",
		"#docu-location EndTransformFeedback http://www.opengl.org/sdk/docs/man4/xhtml/glBeginTransformFeedback.xml",
		"#docu-location EndConditionalRender http://www.opengl.org/sdk/docs/man4/xhtml/glBeginConditionalRender.xml",
		"#docu-location PauseTransformFeedback http://www.opengl.org/sdk/docs/man4/xhtml/glPauseTransformFeedback.xml",
		"#docu-location GetFramebufferAttachmentParameteriv http://www.khronos.org/opengles/sdk/docs/man/xhtml/glGetFramebufferAttachmentParameteriv.xml", // GLES!
		// specify explicit documentation location for some functions

		"#copydoc EnableIndexed Enablei",
		"#copydoc DisableIndexed Disablei",
		"#copydoc IsEnabledIndexed IsEnabledi",
		"#docu-location EnableIndexedEXT http://www.opengl.org/sdk/docs/man4/xhtml/glEnable.xml",
		"#docu-location DisableIndexedEXT http://www.opengl.org/sdk/docs/man4/xhtml/glDisable.xml",
		"#docu-location IsEnabledIndexedEXT http://www.opengl.org/sdk/docs/man4/xhtml/glIsEnabled.xml",
		"#copydoc ColorMaskIndexed ColorMaski",
		"#docu-location ColorMaskIndexedEXT http://www.opengl.org/sdk/docs/man4/xhtml/glColorMask.xml",
		"#copydoc GetBooleanIndexedv GetBooleani_v",
		"#copydoc GetIntegerIndexedv GetIntegeri_v",
		"#copydoc GetIntegerIndexediv GetIntegeri_v",
		"#docu-location GetBooleanIndexedvEXT http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetIntegerIndexedvEXT http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",
		"#docu-location GetIntegerIndexedivEXT http://www.opengl.org/sdk/docs/man4/xhtml/glGet.xml",

		"#deprecate-fun VertexP2ui",
		"#deprecate-fun VertexP3ui",
		"#deprecate-fun VertexP4ui",
		"#deprecate-fun VertexP2uiv",
		"#deprecate-fun VertexP3uiv",
		"#deprecate-fun VertexP4uiv",
		"#deprecate-fun TexCoordP1ui",
		"#deprecate-fun TexCoordP2ui",
		"#deprecate-fun TexCoordP3ui",
		"#deprecate-fun TexCoordP4ui",
		"#deprecate-fun TexCoordP1uiv",
		"#deprecate-fun TexCoordP2uiv",
		"#deprecate-fun TexCoordP3uiv",
		"#deprecate-fun TexCoordP4uiv",
		"#deprecate-fun MultiTexCoordP1ui",
		"#deprecate-fun MultiTexCoordP2ui",
		"#deprecate-fun MultiTexCoordP3ui",
		"#deprecate-fun MultiTexCoordP4ui",
		"#deprecate-fun MultiTexCoordP1uiv",
		"#deprecate-fun MultiTexCoordP2uiv",
		"#deprecate-fun MultiTexCoordP3uiv",
		"#deprecate-fun MultiTexCoordP4uiv",
		"#deprecate-fun NormalP3ui",
		"#deprecate-fun NormalP3uiv",
		"#deprecate-fun ColorP3ui",
		"#deprecate-fun ColorP4ui",
		"#deprecate-fun ColorP3uiv",
		"#deprecate-fun ColorP4uiv",
		"#deprecate-fun SecondaryColorP3ui",
		"#deprecate-fun SecondaryColorP3uiv",
		// deprecate functions manually

		"#gen-alias n_GetARBClEventFuncPointers n_GetARBCLEventFuncPointers",
		"#gen-alias n_GetEXTTextureCompressionLatcFuncPointers n_GetEXTTextureCompressionLAtcFuncPointers",
		"#gen-alias n_GetARBTextureRgb10A2uiFuncPointers n_GetARBTextureRGB10A2uiFuncPointers",
		"#gen-alias n_GetARBVertexArrayBgraFuncPointers n_GetARBVertexArrayBGRAFuncPointers",
		"#gen-alias n_GetARBTextureRgFuncPointers n_GetARBTextureRGFuncPointers",
		"#gen-alias n_GetARBTextureCompressionRgtcFuncPointers n_GetARBTextureCompressionRGtcFuncPointers",
		"#gen-alias n_GetEXTGpuProgramParametersFuncPointers n_GetEXTGPUProgramParametersFuncPointers",
		"#gen-alias n_GetEXTAbgrFuncPointers n_GetEXTABGRFuncPointers",
		"#gen-alias n_GetARBGpuShader5FuncPointers n_GetARBGPUShader5FuncPointers",
		"#gen-alias n_GetARBGpuShaderFp64FuncPointers n_GetARBGPUShaderFP64FuncPointers",
		"#gen-alias n_GetARBTextureBufferObjectRgb32FuncPointers n_GetARBTextureBufferObjectRGB32FuncPointers",
		// fix case in extension entry point retrieval functions

		"#getprocaddr-alias glGetIntegerIndexediv glGetIntegeri_v",
		"#getprocaddr-alias glEnableIndexed glEnablei",
		"#getprocaddr-alias glDisableIndexed glDisablei",
		"#getprocaddr-alias glIsEnabledIndexed glIsEnabledi",
		// fix some functions that are not supported by the driver, in favor of different functions

		"#url-alias http://www.opengl.org/registry/specs/SGIS/fog_function.txt http://www.opengl.org/registry/specs/SGIS/fog_func.txt",
		"#url-alias http://www.opengl.org/registry/specs/SGIX/pixel_texture.txt http://www.opengl.org/registry/specs/SGIX/sgix_pixel_texture.txt",
		"#url-alias http://www.opengl.org/registry/specs/SGIX/texture_add_env.txt http://www.opengl.org/registry/specs/SGIX/texture_env_add.txt",
		"#url-alias http://www.opengl.org/registry/specs/SUN/multi_draw_arrays.txt http://www.opengl.org/registry/specs/EXT/multi_draw_arrays.txt",
		"#url-alias http://www.opengl.org/registry/specs/3DFX/multisample.txt http://www.opengl.org/registry/specs/3DFX/3dfx_multisample.txt",
		"#url-alias http://www.opengl.org/registry/specs/EXT/multisample.txt http://www.opengl.org/registry/specs/EXT/wgl_multisample.txt",
		"#url-alias http://www.opengl.org/registry/specs/SGIX/vertex_preclip_hint.txt http://www.opengl.org/registry/specs/SGIX/vertex_preclip.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/byte_coordinates.txt http://www.opengl.org/registry/specs/OES/OES_byte_coordinates.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/fixed_point.txt http://www.opengl.org/registry/specs/OES/OES_fixed_point.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/single_precision.txt http://www.opengl.org/registry/specs/OES/OES_single_precision.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/compressed_paletted_texture.txt http://www.opengl.org/registry/specs/OES/OES_compressed_paletted_texture.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/read_format.txt http://www.opengl.org/registry/specs/OES/OES_read_format.txt",
		"#url-alias http://www.opengl.org/registry/specs/OES/query_matrix.txt http://www.opengl.org/registry/specs/OES/OES_query_matrix.txt",
		"#url-alias http://www.opengl.org/registry/specs/ARB/vertex_attrib64bit.txt http://www.opengl.org/registry/specs/ARB/vertex_attrib_64bit.txt",
		// some files are stored elsewhere in OpenGL extension registry than one would expect

		"#ext-brand remove",
		"#ext-brand getprocaddress-keep",
		// opengl core functions follow; do not include extension brand
		// in the function name, but keep it in GetProcAddress() string, if present

		/*"#static", // static linking
		"GL_OGL_static_opengl_10_functions",
		"GL_OGL_static_opengl_11_functions",

		"#dynamic", // dynamic linking
		"GL_OGL_static_opengl_12_functions",
		"GL_OGL_static_opengl_13_functions",
		"GL_OGL_static_opengl_14_functions",
		"GL_OGL_static_opengl_15_functions",
		"GL_OGL_static_opengl_20_functions",
		"GL_OGL_static_opengl_21_functions",
		"GL_OGL_static_opengl_30_functions",
		"GL_OGL_static_opengl_31_functions",
		"GL_OGL_static_opengl_32_functions",*/ // make old OpenGL headers (needed once to generate gl??.h)

		"#include src/gl10.h",
		"#include src/gl11.h",
		"#include src/gl12.h",
		"#include src/gl13.h",
		"#include src/gl14.h",
		"#include src/gl15.h",
		"#include src/gl20.h",
		"#include src/gl21.h",
		// include old OpenGL headers which are not extensions themselves

		"#have-def BGRA",
		"#have-def SRGB",
		"#have-def SCISSOR_BOX",
		"#have-def VIEWPORT",
		"#have-def DEPTH_RANGE",
		"#have-def SCISSOR_TEST",
		"#have-def NO_ERROR",
		// avoid some macro redefinitions

		//"#have-fun GetPointerv", // gl11
		"#have-fun BindBufferRange", // gl30
		"#have-fun BindBufferBase", // gl30
		"#have-fun GetIntegeri_v", // gl30
		"#have-fun ProgramParameteri", // gl32
		// avoid some function redefinitions

		"#include src/gl30.h",
		"#comment part of OpenGL 3.0 core",
		"GL_ARB_depth_buffer_float",
		"GL_ARB_framebuffer_object",
		"GL_ARB_framebuffer_sRGB",
		"GL_ARB_half_float_vertex",
		"GL_ARB_map_buffer_range",
		"GL_ARB_texture_compression_rgtc",
		"GL_ARB_texture_rg",
		"GL_ARB_vertex_array_object",
		// OpenGL 3.0

		"#include src/gl31.h",
		"#comment part of OpenGL 3.1 core",
		"GL_ARB_uniform_buffer_object",
		"GL_ARB_draw_instanced",
		"GL_ARB_copy_buffer",
		// OpenGL 3.1

		"#include src/gl32.h",
		"#comment part of OpenGL 3.2 core",
		"GL_ARB_depth_clamp",
		"GL_ARB_draw_elements_base_vertex",
		"GL_ARB_fragment_coord_conventions",
		"GL_ARB_provoking_vertex",
		"GL_ARB_seamless_cube_map",
		"GL_ARB_sync",
		"GL_ARB_texture_multisample",
		"GL_ARB_vertex_array_bgra",
		// OpenGL 3.2

		"#include src/gl33.h",
		"#comment part of OpenGL 3.3 core",
		"GL_ARB_shader_bit_encoding",
		"GL_ARB_blend_func_extended",
		"GL_ARB_explicit_attrib_location",
		"GL_ARB_occlusion_query2",
		"GL_ARB_sampler_objects",
		"GL_ARB_texture_rgb10_a2ui",
		"GL_ARB_texture_swizzle",
		"GL_ARB_timer_query",
		"GL_ARB_instanced_arrays",
		"GL_ARB_vertex_type_2_10_10_10_rev",
		// OpenGL 3.3

		"#include src/gl40.h",
		"#comment part of OpenGL 4.0 core",
		"GL_ARB_texture_query_lod",
		"GL_ARB_draw_buffers_blend",
		"GL_ARB_draw_indirect",
		"GL_ARB_gpu_shader5",
		"GL_ARB_gpu_shader_fp64",
		"GL_ARB_sample_shading",
		"GL_ARB_shader_subroutine",
		"GL_ARB_tessellation_shader",
		"GL_ARB_texture_buffer_object_rgb32",
		"GL_ARB_texture_cube_map_array",
		"GL_ARB_texture_gather",
		"GL_ARB_transform_feedback2",
		"GL_ARB_transform_feedback3",
		// OpenGL 4.0

		"#include src/gl41.h",
		"#comment part of OpenGL 4.1 core",
		"GL_ARB_ES2_compatibility",
		"GL_ARB_get_program_binary",
		"GL_ARB_separate_shader_objects",
		"GL_ARB_shader_precision",
		"GL_ARB_vertex_attrib_64bit",
		"GL_ARB_viewport_array",
		// OpenGL 4.1

		// @note - when adding new OpenGL versions, you need to edit
		// CGL4ExtensionHandler::n_GetOpenGLCoreFuncPointers() and
		// CGL4ExtensionHandler::n_GetAllOpenGLCoreFuncPointers() !!!

		"#ext-brand force",
		"#ext-brand getprocaddress-keep",
		// extension functions follow; force include extension brand
		// in the function name, but keep the GetProcAddress() string as is

		"#comment",
		"GL_EXT_draw_buffers2",
		"GL_EXT_timer_query",
		"#autodoc disable", // disable autodoc, these functions are not yet in OpenGL man
		"GL_EXT_gpu_program_parameters",
		"GL_EXT_bindable_uniform",
		"#autodoc enable", // re-enable autodoc
		"GL_ARB_shader_texture_lod",
		"GL_ARB_texture_buffer_object",
		"GL_S3_s3tc",
		"GL_EXT_texture_compression_latc",
		"GL_EXT_texture_compression_s3tc",
		"GL_ARB_texture_compression_bptc",
		"GL_EXT_texture_filter_anisotropic",
		"GL_EXT_abgr",
		"#autodoc disable", // disable autodoc, these functions are not yet in OpenGL man
		"GL_ARB_shading_language_include",
		"GL_ARB_cl_event",
		"GL_ARB_debug_output",
		"GL_ARB_robustness",
		"#autodoc enable", // re-enable autodoc
		"GL_ARB_shader_stencil_export",
		// just some extensions

		// @note - when adding extensions, you need to edit
		// CGL4ExtensionHandler::n_GetAllSupportedExtensionsFuncPointers() !!!

		/*"GL_ARB_transpose_matrix",
		"GL_ARB_multisample",
		"GL_ARB_texture_env_add",
		"GL_ARB_texture_cube_map",
		"GL_ARB_texture_compression",
		"GL_ARB_texture_border_clamp",
		"GL_ARB_point_parameters",
		"GL_ARB_vertex_blend",
		"GL_ARB_matrix_palette",
		"GL_ARB_texture_env_combine",
		"GL_ARB_texture_env_crossbar",
		"GL_ARB_texture_env_dot3",
		"GL_ARB_texture_mirrored_repeat",
		"GL_ARB_depth_texture",
		"GL_ARB_shadow",
		"GL_ARB_shadow_ambient",
		"GL_ARB_window_pos",
		"GL_ARB_vertex_program",
		"GL_ARB_fragment_program",
		"GL_ARB_vertex_buffer_object",
		"GL_ARB_occlusion_query",
		"GL_ARB_shader_objects",
		"GL_ARB_vertex_shader",
		"GL_ARB_fragment_shader",
		"GL_ARB_shading_language_100",
		"GL_ARB_texture_non_power_of_two",
		"GL_ARB_point_sprite",
		"GL_ARB_fragment_program_shadow",
		"GL_ARB_draw_buffers",
		"GL_ARB_texture_rectangle",
		"GL_ARB_color_buffer_float",
		"GL_ARB_half_float_pixel",
		"GL_ARB_texture_float",
		"GL_ARB_pixel_buffer_object",
		"GL_ARB_depth_buffer_float",
		"GL_ARB_geometry_shader4",*/
		// some extensions i'm not sure about including (some of them are a part of the fixed-function pipeline)

		// vendor / EXT extensions are not included at all (there is a lot of deprecated rubbish i don't really need)
	};
	size_t n_command_num = sizeof(p_command_list) / sizeof(p_command_list[0]);
	// list of extension compiler commands to process

	const char *p_assembly_command_list[] = {
		"#tab-size 4",
		// make four-space tabs (convert spaces to tabs)

		"#assembly ./OpenGL40.h"
			" ./src/head.h"
			" ./tmp/extensions_hdr.h"
			" ./src/head2.h"
			" ./src/autocheck.h"
			" ./tmp/extensions.h"
			" ./src/tail.h",
		// assemble OpenGL40.h

		"#assembly ./OpenGL40.cpp"
			" ./src/head.cpp"
			" ./tmp/extensions.cpp"
			" ./src/tail.cpp",
		// assemble OpenGL40.cpp

		"#tab-size 0",
		// keep spaces instead of tabs

		"#assembly ./gl4.h"
			" ./src/head_nolame.h"
			" ./tmp/extensions_hdr.h"
			" ./src/head2.h"
			" ./src/autocheck.h"
			" ./tmp/extensions.h"
			" ./src/tail.h",
		// assemble gl4.h

		"#assembly ./gl4.cpp"
			" ./src/head_nolame.cpp"
			" ./tmp/extensions.cpp"
			" ./src/tail.cpp"
		// assemble gl4.cpp
	};
	size_t n_assmebly_command_num = sizeof(p_assembly_command_list) / sizeof(p_assembly_command_list[0]);
	// list of final assembly commands to process

	const char *p_s_header_file_name = "./tmp/extensions.h";
	const char *p_s_getters_file_name = "./tmp/extensions_hdr.h";
	const char *p_s_source_file_name = "./tmp/extensions.cpp";
	bool b_want_gl_autocheck = true;
	bool b_force_downloads = false;
	bool b_dont_notify_cached_files = true; // shut up
	// launch configuration

	FILE *p_header_file, *p_getters_file, *p_source_file;
	if(!(p_header_file = fopen(p_s_header_file_name, "w"))) {
		fprintf(stderr, "error: failed to open \'%s\' for writing\n", p_s_header_file_name);
		return -1;
	}
	if(!(p_getters_file = fopen(p_s_getters_file_name, "w"))) {
		fprintf(stderr, "error: failed to open \'%s\' for writing\n", p_s_getters_file_name);
		fclose(p_header_file);
		return -1;
	}
	if(!(p_source_file = fopen(p_s_source_file_name, "w"))) {
		fprintf(stderr, "error: failed to open \'%s\' for writing\n", p_s_source_file_name);
		fclose(p_header_file);
		fclose(p_getters_file);
		return -1;
	}
	// open output files

	if(!Process_Commands(p_command_list, n_command_num, p_header_file, p_getters_file,
	   p_source_file, b_force_downloads, b_dont_notify_cached_files, b_want_gl_autocheck)) {
		fprintf(stderr, "error: Process_Commands() failed\n");
		fclose(p_header_file);
		fclose(p_getters_file);
		fclose(p_source_file);
		return -1;
	}
	// process the commands

	if(ferror(p_header_file) || ferror(p_getters_file) || ferror(p_source_file)) {
		fprintf(stderr, "error: i/o while writing output file(s) (not enough disc space?)\n");
		fclose(p_header_file);
		fclose(p_getters_file);
		fclose(p_source_file);
		return -1;
	}
	fclose(p_header_file);
	fclose(p_getters_file);
	fclose(p_source_file);
	// close output files

	if(!Final_Assembly(p_assembly_command_list, n_assmebly_command_num)) {
		fprintf(stderr, "error: Final_Assembly() failed\n");
		return -1;
	}
	// final assembly step (after the output files are closed)

	return 0;
}

bool Process_Commands(const char **p_command_list, size_t n_command_num,
	FILE *p_header_file, FILE *p_getters_file, FILE *p_source_file,
	bool b_force_downloads, bool b_dont_notify_cached_files,
	bool b_want_gl_autocheck)
{
	bool b_static_linking = false;
	bool b_find_documentation = false;
	const char *p_s_section_comment = "";
	bool b_force_ext_brand_name = false;
	bool b_remove_ext_brand_name = false;
	bool b_force_ext_brand_name_in_getprocaddress_string = false;
	bool b_remove_ext_brand_name_in_getprocaddress_string = false;
	std::map<std::string, std::string> url_translation_table;
	std::map<std::string, std::string> ext_function_name_translation_table;
	std::map<std::string, std::string> entry_point_name_translation_table;
	std::map<std::string, std::string> fun_docu_url_map;
	std::map<std::string, std::string> fun_docu_alias_map;
	// generator configuration

	std::set<std::string> token_alloc;
	std::set<std::string> function_alloc, deprecated_func_set;
	std::set<std::string> extension_set;
	// sets of allocated tokens / functions and of processed extensions

	try {
		for(size_t i = 0; i < n_command_num; ++ i) {
			const char *p_s_command = p_command_list[i];
			// get a command

			if(*p_s_command == '#') {
				if(strstr(p_s_command, "#include") == p_s_command) {
					const char *p_s_include_file = p_s_command + strlen("#include");
					while(*p_s_include_file && isspace(*p_s_include_file))
						++ p_s_include_file;
					// include this file

					FILE *p_fr;
					if(!(p_fr = fopen(p_s_include_file, "r"))) {
						fprintf(stderr, "warning: unable to include \'%s\'\n", p_s_include_file);
						continue;
					}

					std::string s_line;
					while(!feof(p_fr)) {
						if(!stl_ut::ReadLine(s_line, p_fr)) {
							fprintf(stderr, "error: i/o error while reading \'%s\'\n", p_s_include_file);
							fclose(p_fr);
							return false;
						}
						if(s_line.empty() && feof(p_fr))
							break;
						fprintf(p_header_file, "%s\n", s_line.c_str());
					}
					// copy the file to the output

					if(ferror(p_fr)) {
						fprintf(stderr, "error: i/o error while reading \'%s\'\n", p_s_include_file);
						fclose(p_fr);
						return false;
					}
					fclose(p_fr);
				} else if(strstr(p_s_command, "#static") == p_s_command) {
					b_static_linking = true;
				} else if(strstr(p_s_command, "#dynamic") == p_s_command) {
					b_static_linking = false;
				} else if(strstr(p_s_command, "#comment") == p_s_command) {
					p_s_section_comment = p_s_command + strlen("#comment");
					while(*p_s_section_comment && isspace(*p_s_section_comment))
						++ p_s_section_comment;
				} else if(strstr(p_s_command, "#have-def") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#have-def");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "GL_") == p_s_name)
						p_s_name += 3;
					token_alloc.insert(std::string(p_s_name));
				} else if(strstr(p_s_command, "#have-fun") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#have-fun");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "gl") == p_s_name)
						p_s_name += 2;
					function_alloc.insert(std::string(p_s_name));
				} else if(strstr(p_s_command, "#deprecate-fun") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#deprecate-fun");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "gl") == p_s_name)
						p_s_name += 2;
					deprecated_func_set.insert(std::string(p_s_name));
				} else if(strstr(p_s_command, "#autodoc") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#autodoc");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(!strcmp(p_s_name, "enable"))
						b_find_documentation = true;
					else if(!strcmp(p_s_name, "disable"))
						b_find_documentation = false;
					else {
						fprintf(stderr, "error: unknown command token: \'%s\'\n", p_s_name);
						return false;
					}
				} else if(strstr(p_s_command, "#gen-alias") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#gen-alias");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_end = p_s_name;
					std::string s_name, s_alias;
					s_name.insert(s_name.begin(), p_s_first_name_begin, p_s_first_name_end);
					s_alias.insert(s_alias.begin(), p_s_second_name_begin, p_s_second_name_end);
					ext_function_name_translation_table[s_name] = s_alias;
				} else if(strstr(p_s_command, "#getprocaddr-alias") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#getprocaddr-alias");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_end = p_s_name;
					std::string s_name, s_alias;
					s_name.insert(s_name.begin(), p_s_first_name_begin, p_s_first_name_end);
					s_alias.insert(s_alias.begin(), p_s_second_name_begin, p_s_second_name_end);
					entry_point_name_translation_table[s_name] = s_alias;
				} else if(strstr(p_s_command, "#url-alias") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#url-alias");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_end = p_s_name;
					std::string s_url, s_alias;
					s_url.insert(s_url.begin(), p_s_first_name_begin, p_s_first_name_end);
					s_alias.insert(s_alias.begin(), p_s_second_name_begin, p_s_second_name_end);
					url_translation_table[s_url] = s_alias;
				} else if(strstr(p_s_command, "#copydoc") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#copydoc");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "gl") == p_s_name)
						p_s_name += 2;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "gl") == p_s_name)
						p_s_name += 2;
					const char *p_s_second_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_end = p_s_name;
					std::string s_name, s_alias;
					s_name.insert(s_name.begin(), p_s_first_name_begin, p_s_first_name_end);
					s_alias.insert(s_alias.begin(), p_s_second_name_begin, p_s_second_name_end);
					fun_docu_alias_map[s_name] = s_alias;
				} else if(strstr(p_s_command, "#docu-location") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#docu-location");
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					if(strstr(p_s_name, "gl") == p_s_name)
						p_s_name += 2;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_second_name_end = p_s_name;
					std::string s_name, s_alias;
					s_name.insert(s_name.begin(), p_s_first_name_begin, p_s_first_name_end);
					s_alias.insert(s_alias.begin(), p_s_second_name_begin, p_s_second_name_end);
					fun_docu_url_map[s_name] = s_alias;
				} else if(strstr(p_s_command, "#ext-brand") == p_s_command) {
					const char *p_s_option = p_s_command + strlen("#ext-brand");
					while(*p_s_option && isspace(*p_s_option))
						++ p_s_option;
					if(!strcmp(p_s_option, "remove")) {
						b_force_ext_brand_name = false;
						b_remove_ext_brand_name = true;
					} else if(!strcmp(p_s_option, "keep")) {
						b_force_ext_brand_name = false;
						b_remove_ext_brand_name = false;
					} else if(!strcmp(p_s_option, "force")) {
						b_force_ext_brand_name = true;
						b_remove_ext_brand_name = false;
					} else if(!strcmp(p_s_option, "getprocaddress-remove")) {
						b_force_ext_brand_name_in_getprocaddress_string = false;
						b_remove_ext_brand_name_in_getprocaddress_string = true;
					} else if(!strcmp(p_s_option, "getprocaddress-keep")) {
						b_force_ext_brand_name_in_getprocaddress_string = false;
						b_remove_ext_brand_name_in_getprocaddress_string = false;
					} else if(!strcmp(p_s_option, "getprocaddress-force")) {
						b_force_ext_brand_name_in_getprocaddress_string = true;
						b_remove_ext_brand_name_in_getprocaddress_string = false;
					} else {
						fprintf(stderr, "error: unknown command token: \'%s\'\n", p_s_option);
						return false;
					}
					// extensions brand handling
				} else if(strstr(p_s_command, "#assembly") == p_s_command) {
					const char *p_s_name = p_s_command + strlen("#assembly");

					std::string s_outfile;
					{
						while(*p_s_name && isspace(*p_s_name))
							++ p_s_name;
						const char *p_s_first_name_begin = p_s_name;
						while(*p_s_name && !isspace(*p_s_name))
							++ p_s_name;
						const char *p_s_first_name_end = p_s_name;
						s_outfile.insert(s_outfile.begin(), p_s_first_name_begin, p_s_first_name_end);
					}
					// get output file

					FILE *p_assembled_file;
					if(!(p_assembled_file = fopen(s_outfile.c_str(), "w"))) {
						fprintf(stderr, "error: failed to create output file \'%s\'\n", s_outfile.c_str());
						return false;
					}
					// open file being assembled

					for(;;) {
						while(*p_s_name && isspace(*p_s_name))
							++ p_s_name;
						const char *p_s_next_name_begin = p_s_name;
						while(*p_s_name && !isspace(*p_s_name))
							++ p_s_name;
						const char *p_s_next_name_end = p_s_name;
						if(p_s_next_name_begin == p_s_next_name_end)
							break;
						std::string s_append_file;
						s_append_file.insert(s_append_file.begin(), p_s_next_name_begin, p_s_next_name_end);
						// get file to append

						FILE *p_fr;
						if(!(p_fr = fopen(s_append_file.c_str(), "r"))) {
							fprintf(stderr, "error: unable to open \'%s\'\n", s_append_file.c_str());
							fclose(p_assembled_file);
							return false;
						}
						// open appended file

						std::string s_line;
						while(!feof(p_fr)) {
							if(!stl_ut::ReadLine(s_line, p_fr)) {
								fprintf(stderr, "error: i/o error while reading \'%s\'\n", s_append_file.c_str());
								fclose(p_fr);
								fclose(p_assembled_file);
								return false;
							}
							fprintf(p_assembled_file, "%s\n", s_line.c_str());
						}
						// copy the file to the output

						if(ferror(p_fr)) {
							fprintf(stderr, "error: i/o error while reading \'%s\'\n", s_append_file.c_str());
							fclose(p_fr);
							fclose(p_assembled_file);
							return false;
						}
						fclose(p_fr);
						// close appended file
					}
					// append various files

					if(ferror(p_assembled_file)) {
						fclose(p_assembled_file);
						fprintf(stderr, "error: i/o error while writing \'%s\'\n", s_outfile.c_str());
						return false;
					}
					fclose(p_assembled_file);
					// close
				} else {
					// unknown command
					fprintf(stderr, "error: unknown command: \'%s\'\n", p_s_command);
					return false;
				}

				continue;
			}
			// handle commands

			const char *p_s_ext = p_s_command;
			// otherwise it's an extension name

			if(extension_set.find(std::string(p_s_ext)) != extension_set.end()) {
				fprintf(stderr, "warning: skipping %s extension : have this one already\n", p_s_ext);
				continue;
			}
			extension_set.insert(std::string(p_s_ext));
			// keep a set of extensions that were processed already

			if(strstr(p_s_ext, "GL_") != p_s_ext) {
				fprintf(stderr, "warning: skipping %s extension : not a GL extension\n", p_s_ext);
				continue;
			}
			// skip non-gl extensions

			const char *p_s_vendor = p_s_ext + 3;
			const char *p_s_vendor_end = strchr(p_s_ext + 3, '_');
			const char *p_s_ext_name = p_s_vendor_end + 1;
			if(!p_s_vendor_end) {
				fprintf(stderr, "warning: skipping %s extension : unable to determine vendor\n", p_s_ext);
				continue;
			}
			std::string s_brand, s_name;
			s_name = p_s_ext_name;
			s_brand.insert(s_brand.begin(), p_s_vendor, p_s_vendor_end);
			// get brand / name

			std::string s_url, s_file;
			{
				s_url = "http://www.opengl.org/registry/specs/";
				s_url += s_brand;
				s_url += "/";
				s_url += s_name;
				s_url += ".txt";

				if(url_translation_table.find(s_url.c_str()) != url_translation_table.end())
					s_url = url_translation_table[s_url.c_str()];
				// translate URL's

				s_file = ".\\ext\\";
				s_file += p_s_ext;
				s_file += ".txt";
			}
			// get URL to extension on OpenGL.org, and filename to download it to

			if(extension_set.find(std::string(s_url.c_str())) != extension_set.end()) {
				fprintf(stderr, "warning: skipping %s extension : "
					"have this one already from \'%s\'\n", p_s_ext, s_url.c_str());
				continue;
			}
			extension_set.insert(std::string(s_url.c_str()));
			// keep a set of extensions that were processed already

			if(!b_force_downloads && !_access(s_file.c_str(), 0)) {
				if(!b_dont_notify_cached_files)
					printf("not downloading \'%s\', it is cached\n", s_url.c_str());
				// verbose
			} else {
				printf("downloading \'%s\' ...\n", s_url.c_str());
				// verbose

				CDownloader downloader;
				int n_result;
				if(!downloader.StartDownload(s_file.c_str(), s_url.c_str()) ||
				   (n_result = downloader.n_WaitForFinish()) == down_Failure) {
					fprintf(stderr, "error: failed to download \'%s\' (fatal error)\n", s_url.c_str());
					return false;
				}
				if(n_result == down_FileDoesntExist) {
					fprintf(stderr, "error: failed to download \'%s\' (404)\n", s_url.c_str());
					continue;
				}
				_ASSERTE(n_result == down_Success);
			}
			// download the file

			CGLExtension ext(p_s_ext);
			if(!ext.Parse_FromFile(s_file.c_str(), deprecated_func_set,
			   b_find_documentation, fun_docu_url_map, fun_docu_alias_map))
				fprintf(stderr, "error: Parse_FromFile() failed on \'%s\'\n", s_file.c_str());
			else {
				if(b_static_linking) {
					ext.Print_StaticHeader(p_header_file, b_want_gl_autocheck, p_s_section_comment,
						b_force_ext_brand_name, b_remove_ext_brand_name, token_alloc,
						function_alloc, deprecated_func_set);
				} else {
					ext.Print_Header(p_header_file, b_want_gl_autocheck, p_s_section_comment,
						b_force_ext_brand_name, b_remove_ext_brand_name, token_alloc,
						function_alloc, deprecated_func_set);
				}
				ext.Print_Header2(p_getters_file, p_s_section_comment,
					ext_function_name_translation_table);
				if(!b_static_linking) {
					ext.Print_Source(p_source_file, p_s_section_comment,
						b_force_ext_brand_name, b_remove_ext_brand_name,
						b_force_ext_brand_name_in_getprocaddress_string,
						b_remove_ext_brand_name_in_getprocaddress_string, function_alloc,
						ext_function_name_translation_table, deprecated_func_set,
						entry_point_name_translation_table);
				}
				// print all the headers

				ext.MarkAllocatedTokens(token_alloc);
				ext.MarkAllocatedFunctions(function_alloc);
				// mark allocated tokens and functions so they are not redefined
			}
			// parse it, write headers
		}
		// loop trough commands
	} catch(std::bad_alloc&) {
		fprintf(stderr, "error: not enough memory\n");
		return false;
	}

	return true;
}

bool Final_Assembly(const char **p_command_list, size_t n_command_num)
{
	int n_tab_size = 0;
	// generator configuration

	try {
		for(size_t i = 0; i < n_command_num; ++ i) {
			const char *p_s_command = p_command_list[i];
			// get a command

			if(strstr(p_s_command, "#tab-size") == p_s_command) {
				const char *p_s_size = p_s_command + strlen("#tab-size");

				while(*p_s_size && isspace(*p_s_size))
					++ p_s_size;
				if(!*p_s_size || !isdigit(*p_s_size)) {
					fprintf(stderr, "error: #tab-size must be followed by a (non-negative) number\n");
					return false;
				}
				int n_new_tab_size = atoi(p_s_size);
				if(n_new_tab_size < 0) {
					fprintf(stderr, "warning: invalid tab-size: %d\n", n_new_tab_size);
					continue;
				} else
					n_tab_size = n_new_tab_size;
				// handle tab-size command
			} else if(strstr(p_s_command, "#assembly") == p_s_command) {
				const char *p_s_name = p_s_command + strlen("#assembly");

				std::string s_outfile;
				{
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_first_name_end = p_s_name;
					s_outfile.insert(s_outfile.begin(), p_s_first_name_begin, p_s_first_name_end);
				}
				// get output file

				FILE *p_assembled_file;
				if(!(p_assembled_file = fopen(s_outfile.c_str(), "w"))) {
					fprintf(stderr, "error: failed to create output file \'%s\'\n", s_outfile.c_str());
					return false;
				}
				// open file being assembled

				for(;;) {
					while(*p_s_name && isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_next_name_begin = p_s_name;
					while(*p_s_name && !isspace(*p_s_name))
						++ p_s_name;
					const char *p_s_next_name_end = p_s_name;
					if(p_s_next_name_begin == p_s_next_name_end)
						break;
					std::string s_append_file;
					try {
						s_append_file.insert(s_append_file.begin(), p_s_next_name_begin, p_s_next_name_end);
					} catch(std::bad_alloc&) {
						fclose(p_assembled_file); // close the file ...
						throw std::bad_alloc(); // ... and rethrow
					}
					// get name of file to append

					FILE *p_fr;
					if(!(p_fr = fopen(s_append_file.c_str(), "r"))) {
						fprintf(stderr, "error: unable to open \'%s\'\n", s_append_file.c_str());
						fclose(p_assembled_file);
						return false;
					}
					// open appended file

					std::string s_line;
					while(!feof(p_fr)) {
						if(!stl_ut::ReadLine(s_line, p_fr)) {
							fprintf(stderr, "error: i/o error while reading \'%s\'\n", s_append_file.c_str());
							fclose(p_fr);
							fclose(p_assembled_file);
							return false;
						}
						if(s_line.empty() && feof(p_fr))
							break;

						if(n_tab_size) {
							_ASSERTE(n_tab_size > 0);

							for(size_t i = 0, n_modulo_pos = 0; i < s_line.length(); ++ i) {
								int n_char = (unsigned char)s_line[i];
								if(n_char == ' ' && n_modulo_pos == n_tab_size - 1) {
									size_t n_first_space = i;
									while(n_first_space - 1 >= 0 && i - n_first_space + 1 < unsigned(n_tab_size) &&
									   s_line[n_first_space - 1] == ' ')
										-- n_first_space;
									// find the first space

									if(i == n_first_space && (i + 1 == s_line.length() || s_line[i + 1] != ' ')) {
										++ n_modulo_pos;
										// in case there is just a single space (not followed by more,
										// which could be translated to tabs), don't replace it 
									} else {
										_ASSERTE(i - n_first_space + 1 <= unsigned(n_tab_size));
										// make sure there are not too many spaces

										s_line.erase(n_first_space + 1, i - n_first_space);
										s_line[n_first_space] = '\t';
										// replace the spaces by a tab

										i = n_first_space;
										// move back to the start of the space sequence (mind the ++ i)

										n_modulo_pos = 0;
										// just had a tab, the position is even multiple of n_tab_size
									}
								} else if(n_char == '\t')
									n_modulo_pos = n_Align_Up(n_modulo_pos + 1, size_t(n_tab_size)); // tab aligns position up to the next tab size
								else
									++ n_modulo_pos; // ordinary characters just increment position

								if(n_modulo_pos == n_tab_size)
									n_modulo_pos = 0;
								_ASSERTE(n_modulo_pos < unsigned(n_tab_size));
								// modulo operation
							}
						}

						fprintf(p_assembled_file, "%s\n", s_line.c_str());
					}
					// copy the file to the output

					if(ferror(p_fr)) {
						fprintf(stderr, "error: i/o error while reading \'%s\'\n", s_append_file.c_str());
						fclose(p_fr);
						fclose(p_assembled_file);
						return false;
					}
					fclose(p_fr);
					// close appended file
				}
				// append various files

				if(ferror(p_assembled_file)) {
					fclose(p_assembled_file);
					fprintf(stderr, "error: i/o error while writing \'%s\'\n", s_outfile.c_str());
					return false;
				}
				fclose(p_assembled_file);
				// close
			} else {
				// unknown command
				fprintf(stderr, "error: unknown command: \'%s\'\n", p_s_command);
				return false;
			}
			// handle commands
		}
	} catch(std::bad_alloc&) {
		fprintf(stderr, "error: not enough memory\n");
		return false;
	}

	return true;
}
