/*
                                +--------------------------------+
                                |                                |
                                |     ***   OpenGL 4.0   ***     |
                                |                                |
                                |  Copyright  -tHE SWINe- 2011  |
                                |                                |
                                |           OpenGL40.h           |
                                |                                |
                                +--------------------------------+
*/

#ifndef __OPENGL40_TOOL_INCLUDED
#define __OPENGL40_TOOL_INCLUDED

/**
 *  @file gl4/OpenGL40.h
 *  @author -tHE SWINe-
 *  @brief OpenGL 4.0 and extensions entry points
 *
 *  This header contains entry points of relevant OpenGL 4 functions / extensions.
 *      It was prepared from gl3.h, available from <www.opengl.org>,
 *      from extension registry at <www.opengl.org> and from OpenGL function
 *      documentation pages at <http://www.opengl.org/sdk/docs/man4/xhtml/>.
 *
 *  The difference between OpenGL30.h and OpenGL40.h is OpenGL40.h also includes
 *      OpenCL headers to support OpenCL interop (GL_ARB_cl_event).
 *
 *  @date 2009-11-28
 *
 *  core functionality / core extensions:<br/>
 *      OpenGL 1.0<br/>
 *      OpenGL 1.1<br/>
 *      OpenGL 1.2<br/>
 *      OpenGL 1.3<br/>
 *      OpenGL 1.4<br/>
 *      OpenGL 1.5<br/>
 *      OpenGL 2.0<br/>
 *      OpenGL 2.1<br/>
 *      OpenGL 3.0
 *          - GL_ARB_depth_buffer_float
 *          - GL_ARB_framebuffer_object
 *          - GL_ARB_framebuffer_sRGB
 *          - GL_ARB_half_float_vertex
 *          - GL_ARB_map_buffer_range
 *          - GL_ARB_texture_compression_rgtc
 *          - GL_ARB_texture_rg
 *          - GL_ARB_vertex_array_object
 *
 *      OpenGL 3.1
 *          - GL_ARB_uniform_buffer_object
 *          - GL_ARB_draw_instanced
 *          - GL_ARB_copy_buffer
 *
 *      OpenGL 3.2
 *          - GL_ARB_depth_clamp
 *          - GL_ARB_draw_elements_base_vertex
 *          - GL_ARB_fragment_coord_conventions
 *          - GL_ARB_provoking_vertex
 *          - GL_ARB_seamless_cube_map
 *          - GL_ARB_sync
 *          - GL_ARB_texture_multisample
 *          - GL_ARB_vertex_array_bgra
 *
 *  present extensions:
 *          - GL_EXT_draw_buffers2          (separate blend enables, separate color masks for mutliple draw buffers)
 *          - GL_ARB_draw_buffers_blend     (different blend equations for multiple draw buffers)
 *          - GL_ARB_sample_shading         (enforced shading for multisample antialiassing)
 *          - GL_ARB_texture_cube_map_array (texture array for cube-maps)
 *          - GL_ARB_texture_gather         (new functions for GLSL, returns x components of four sampled texels, lets shader do it's own filtering, instead of bilinear. without GL_EXT_texture_swizzle it's only monochrome though)
 *          - GL_ARB_texture_query_lod      (new functions for GLSL, returns LOD sampler would use for texture (instead of sampled color))
 *          - GL_EXT_timer_query                (new query object target for GPU timing)
 *          - GL_EXT_gpu_program_parameters (single call to fill multiple GPU program parameters; useless because we do not have programs anymore - only shaders)
 *          - GL_EXT_bindable_uniform           (provides bindable buffers to contain uniform variables, so they don't have to be specified)
 *          - GL_ARB_instanced_arrays           (sets divisor to vertex attributes; when set to n = nonzero, a particular attribute advances every n vertices; similar to direct-x function SetStreamFrequency)
 *          - GL_ARB_shader_texture_lod     (additions to GLSL; contains no new tokens for GL)
 *          - GL_ARB_texture_buffer_object  (new texture type, buffer object is now bindable as large 1D texture; it's unclear how exactly is it used)
 *
 *          - GL_S3_s3tc                        (pixel formats)
 *          - GL_EXT_texture_compression_latc   (pixel formats)
 *          - GL_EXT_texture_compression_s3tc   (pixel formats)
 *          - GL_EXT_texture_filter_anisotropic (anisotropic texture filter control)
 *          - GL_EXT_abgr                       (ABGR pixel format)
 *
 *  new extensions which are not present yet:
 *      -   GL_ARB_compatibility - a big one, fork of OpenGL 3.1
 *
 *  removed:
 *          - GL_ARB_texture_compression - promoted to core in OpenGL 1.3
 *          - GL_ARB_vertex_blend - relies on deprecated parts of fixed-function pipeline
 *          - GL_ARB_matrix_palette - relies on deprecated parts of fixed-function pipeline
 *          - GL_EXT_compiled_vertex_array - is now obsolete, extension wasn't updated for a long time
 *          - GL_EXT_vertex_array - is obsolete, became core and in turn, became deprecated
 *          - GL_ARB_vertex_program, GL_ARB_fragment_program - deprecated; only shading language 1.30 is supported
 *          - GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shader_object - became core in OpenGL 2.0
 *          - GL_ARB_vertex_buffer_object - became core in OpenGL 1.5
 *          - GL_ARB_pixel_buffer_object - became core in OpenGL 2.1
 *          - GL_ARB_occlusion_query - became core in OpenGL 1.5
 *          - GL_ARB_multitexture - became core in OpenGL 1.3; n_GetMultitextureFuncPointers() was merged into n_GetGL13FuncPointers()
 *          - GL_EXT_packed_depth_stencil - part of GL_ARB_framebuffer_object, which is core in OpenGL 3.0
 *          - GL_EXT_framebuffer_object - part of GL_ARB_framebuffer_object, which is core in OpenGL 3.0
 *          - GL_EXT_draw_instanced - became core in OpenGL 3.1
 *
 *  not present:
 *          - GL_ARB_imaging (part of OpenGL 1.2 core) - seems to be deprecated
 *          - GL_ARB_color_buffer_float - superseded by OpenGL 3.0
 *          - GL_NV_fence - superseded by GL_ARB_sync
 *          - GL_EXT_texture_integer - superseded by OpenGL 3.0
 *          - GL_EXT_gpu_shader4 - superseded by OpenGL 3.0, OpenGL 3.1
 *          - GL_EXT_geometry_shader4 - superseded by GL_ARB_framebuffer_object, OpenGL 3.2
 *          - GL_EXT_texture_array - superseded by GL_ARB_framebuffer_object, OpenGL 3.0
 *          - GL_EXT_texture_buffer_object - superseded by OpenGL 3.1
 *          - GL_EXT_blend_minmax - superseded by GL_ARB_imaging, OpenGL 1.2
 *          - GL_EXT_blend_subtract - superseded by GL_ARB_imaging, OpenGL 1.2
 *          - GL_ARB_geometry_shader4 - superseded by GL_ARB_framebuffer_object, OpenGL 3.2
 *          - GL_ARB_draw_instanced - directly part of OpenGL 3.1 core
 *          - GL_ARB_half_float_pixel - superseded by GL_ARB_half_float_vertex, part of OpenGL 3.0 core
 *          - GL_ARB_texture_float - part of OpenGL 3.0 core
 *          - GL_ARB_pixel_buffer_object - part of OpenGL 2.1 core
 *
 *  @date 2010-09-19
 *
 *  added the following extensions:
 *          - GL_ARB_shading_language_include
 *          - GL_ARB_texture_compression_bptc
 *          - GL_ARB_blend_func_extended
 *          - GL_ARB_explicit_attrib_location
 *          - GL_ARB_occlusion_query2
 *          - GL_ARB_sampler_objects
 *          - GL_ARB_shader_bit_encoding
 *          - GL_ARB_texture_rgb10_a2ui
 *          - GL_ARB_texture_swizzle
 *          - GL_ARB_timer_query
 *          - GL_ARB_vertex_type_2_10_10_10_rev
 *          - GL_ARB_draw_indirect
 *          - GL_ARB_gpu_shader5
 *          - GL_ARB_gpu_shader_fp64
 *          - GL_ARB_shader_subroutine
 *          - GL_ARB_tessellation_shader
 *          - GL_ARB_texture_buffer_object_rgb32
 *          - GL_ARB_transform_feedback2
 *          - GL_ARB_transform_feedback3
 *          - GL_ARB_es2_compatibility
 *          - GL_ARB_get_program_binary
 *          - GL_ARB_separate_shader_objects
 *          - GL_ARB_shader_precision
 *          - GL_ARB_vertex_attrib64bit
 *          - GL_ARB_viewport_array
 *          - GL_ARB_cl_event
 *          - GL_ARB_debug_output
 *          - GL_ARB_robustness
 *          - GL_ARB_shader_stencil_export
 *
 *  @date 2010-10-29
 *
 *  Unified windows detection macro to "#if defined(_WIN32) || defined(_WIN64)".
 *
 *  @date 2011-07-15
 *
 *  Removed glFramebufferTextureFace(), which was a bug in the documentation. OpenGL 3.2 support
 *  is now reported complete (CGL3ExtensionHandler::n_GetGL32FuncPointers() returns 0,
 *  and not 1 as it did).
 *
 *  Re-organized core functionality (the extensions were all present already):
 *
 *      OpenGL 3.3:
 *          - GL_ARB_shader_bit_encoding
 *          - GL_ARB_blend_func_extended
 *          - GL_ARB_explicit_attrib_location
 *          - GL_ARB_occlusion_query2
 *          - GL_ARB_sampler_objects
 *          - GL_ARB_texture_rgb10_a2ui
 *          - GL_ARB_texture_swizzle
 *          - GL_ARB_timer_query
 *          - GL_ARB_instanced_arrays
 *          - GL_ARB_vertex_type_2_10_10_10_rev
 *
 *      OpenGL 4.0:
 *          - GL_ARB_texture_query_lod
 *          - GL_ARB_draw_indirect
 *          - GL_ARB_gpu_shader5
 *          - GL_ARB_gpu_shader_fp64
 *          - GL_ARB_shader_subroutine
 *          - GL_ARB_tessellation_shader
 *          - GL_ARB_texture_buffer_object_rgb32
 *          - GL_ARB_texture_cube_map_array
 *          - GL_ARB_texture_gather
 *          - GL_ARB_transform_feedback2
 *          - GL_ARB_transform_feedback3
 *
 *      OpenGL 4.1:
 *          - ARB_ES2_compatibility
 *          - ARB_get_program_binary
 *          - ARB_separate_shader_objects
 *          - ARB_shader_precision
 *          - ARB_vertex_attrib_64bit
 *          - ARB_viewport_array
 *
 *  Updated the license information.
 *
 *  @date 2011-07-20
 *
 *  Improved autocheck for statically-linked void functions so if using the user breakpoint,
 *  the code displayed in MSVC is user code (the call that caused the error) and not
 *  OpenGL40.h. This only increases user comfort, the functionality didn't change.
 *
 *	Optimized the code to reduce compilation time.
 *
 */

// license from gl3.h
/*
** Copyright (c) 2007-2011 The Khronos Group Inc.
**
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and/or associated documentation files (the
** "Materials"), to deal in the Materials without restriction, including
** without limitation the rights to use, copy, modify, merge, publish,
** distribute, sublicense, and/or sell copies of the Materials, and to
** permit persons to whom the Materials are furnished to do so, subject to
** the following conditions:
**
** The above copyright notice and this permission notice shall be included
** in all copies or substantial portions of the Materials.
**
** THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
** MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
** TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
** MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
*/

// license for materials from http://www.opengl.org/sdk/docs/man/xhtml/
/*
** SGI FREE SOFTWARE LICENSE B (Version 2.0, Sept. 18, 2008)
** Copyright (C) 1991-2006 Silicon Graphics, Inc. All Rights Reserved.
** Permission is hereby granted, free of charge, to any person obtaining
** a copy of this software and associated documentation files (the "Software"),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, sublicense,
** and/or sell copies of the Software, and to permit persons to whom the Software
** is furnished to do so, subject to the following conditions:
**
** The above copyright notice including the dates of first publication and
** either this permission notice or a reference to http://oss.sgi.com/projects/FreeB/
** shall be included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
** OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
** SILICON GRAPHICS, INC. BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
** OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
**
** Except as contained in this notice, the name of Silicon Graphics, Inc. shall
** not be used in advertising or otherwise to promote the sale, use or other
** dealings in this Software without prior written authorization from
** Silicon Graphics, Inc.
*/

#include <string>
#include <stdio.h>
#include "../Integer.h"
#include "../StlUtils.h"
#if defined(_WIN32) || defined (_WIN64)
#define WIN32_LEAND_AND_MEAN 1
#include <windows.h> // need APIENTRY
#endif // _WIN32, _WIN64

/*
 *  makes life a little bit easier when including gl.h by mistake
 *
 *  does not include gl.h, defines GL up to 1.1 below, want to prevent gl.h
 *  header inclusion
 */
#if defined(__gl_h_) || defined(__GL_H__) || defined(GL_VERSION_1_1)
#error("warning: gl.h must NOT be included in files using OpenGL40.h")
#endif

/*
 *  makes life a little bit easier when including glext.h by mistake
 *
 *  glext.h is not included anymore, it does contain significant amount of
 *  deprecated functionality, want to prevent glext.h header inclusion
 */
#if defined(__glext_h_) || defined(GL_GLEXT_VERSION)
#error("glext.h must NOT be included in files using OpenGL40.h")
#endif

/**
 *  @def __gl_h_
 *  @brief avoids inclusion of gl.h
 */
#define __gl_h_

/**
 *  @def __GL_H__
 *  @brief avoids inclusion of gl.h
 */
#define __GL_H__

/**
 *  @def __glext_h_
 *  @brief avoids inclusion of glext.h
 */
#define __glext_h_

/**
 *  @def __GLEXT_H__
 *  @brief avoids inclusion of glext.h
 */
#define __GLEXT_H__

/**
 *  @def NULL_FUNCTION_POINTERS
 *  @brief if defined, initializes entry function pointers to null
 *
 *  In case NULL_FUNCTION_POINTERS is defined (default), all extension function pointers
 *      are initialized to 0, but user never needs to compare individual function values
 *      so it's posible to save some space in executable file by skipping the initialization.
 */
#define NULL_FUNCTION_POINTERS

/**
 *  @def FIX_GL_FUNCTION_LINKING_CONFLICTS
 *  @brief avoids function linking conflicts by renaming function pointers, and fixing
 *      names using #define
 *
 *  Some compilers may contain newer OpenGL libraries than this header requires
 *      (OpenGL 1.1, as supported by Microsoft Visual studio family), so names of function
 *      pointers, such as <tt>glBlendColor</tt> are present in static library, as well as in this
 *      file, causing link-time error. This is fixed by renaming the function, here
 *      <tt>glBlendColor_ptr</tt> would be used, and fixing it by macro, such as <tt>#define
 *      glBlendColor_ptr glBlendColor</tt>.
 *
 *	@note This must not be changed for files generated with the EXPAND_GLFuncPtrName_MACRO option.
 */
#define FIX_GL_FUNCTION_LINKING_CONFLICTS

/**
 *  @def GL_AUTOCHECK
 *  @brief automatically checks for OpenGL errors after every OpenGL function except glGetError()
 *  @note This is only available in debug mode unless FORCE_GL_CHECK is defined.
 */
#define GL_AUTOCHECK

/**
 *  @def GL_AUTOCHECK_USE_EXCEPTION
 *  @brief if defined, autocheck throws std::runtime_exception if an OpenGL error occurs
 */
//#define GL_AUTOCHECK_USE_EXCEPTION

/**
 *  @def GL_AUTOCHECK_USE_BREAKPOINT
 *  @brief if defined, autocheck calls user breakpoint if an OpenGL error occurs
 *  @note This is only available in windows / MSVC and in x86 mode (not x64).
 */
#define GL_AUTOCHECK_USE_BREAKPOINT

/**
 *  @def GL_AUTOCHECK_USE_ABORT
 *  @brief if defined, autocheck calls clib's abort() if an OpenGL error occurs
 */
//#define GL_AUTOCHECK_USE_ABORT

/**
 *  @def GL_AUTOCHECK_USE_EXIT
 *  @brief if defined, autocheck prints the error to stderr and calls
 *      exit(-1) if an OpenGL error occurs
 *  @note This is default behavior if nothing else is defined.
 */
//#define GL_AUTOCHECK_USE_EXIT

/**
 *  @brief class for handling OpenGL extensions
 *
 *  Supplies means to get OpenGL function entry point, check extension support, or get
 *  OpenGL version. Functions with name n_Get*FuncPointers() returns number of function
 *  pointers that were not found (zero is success).
 */
class CGL4ExtensionHandler {
public:
    /**
     *  @brief compares OpenGL version to reference
     *
     *  Compares supplied version to current OpenGL implementation version and returns
     *      true if current version is greater or equal (supplied version is supported).
     *
     *  @param[in] n_min_major is reference minimal major version number
     *  @param[in] n_min_minor is reference minimal minor version number
     *
     *  @return Returns true in case OpenGL version is greater or equal to n_min_major.n_min_minor.
     */
    static bool b_Support_OpenGL(int n_min_major = 4, int n_min_minor = 0);

    /**
     *  @brief returns OpenGL version
     *
     *  Returns OpenGL version as (comparable) integer value.
     *
     *  @return Returns OpenGL version as major * 10 + minor (ie. for OpenGL 3.2 returns 32).
     */
    static int n_OpenGL_Version();

    /**
     *  @brief returns OpenGL version major
     *
     *  @return Returns OpenGL version as major (ie. for OpenGL 3.2 returns 3).
     */
    static int n_OpenGL_Version_Major();

    /**
     *  @brief returns OpenGL version minor
     *
     *  @return Returns OpenGL version as minor (ie. for OpenGL 3.2 returns 2).
     */
    static int n_OpenGL_Version_Minor();

    /**
     *  @brief checks extension support
     *
     *  Checks support of a given OpenGL extension, returns true if extension is supported.
     *
     *  @param[in] p_s_extension_name is null-terminated string, containing name
     *      of OpenGL extension encoded in UTF-8
     *
     *  @return Returns true in case p_s_extension_name is supported, otherwise false.
     */
    static bool b_SupportedExtension(const char *p_s_extension_name);

    /**
     *  @brief gets extension string
     *
     *  OpenGL 3.0 doesn't have classic extension string, it returns names of individual
     *      extensions instead. This is to prevent buffer overruns, people always seem
     *      to be copying OpenGL extensions to fixed-size buffers, causing older software
     *      from times when extension string was shorter to fail on modern hardware.
     *      This functions returns "classic" OpenGL extension string in std::string,
     *      encoded in UTF-8, names of individual extensions are separated by space (' ')
     *      character.
     *
     *  @return Returns true on success, false on failure (not enough memory).
     */
    static bool Get_ExtensionString(std::string &r_s_extension_list);

#if defined(_WIN32) || defined (_WIN64)
    /**
     *  @brief checks extension support
     *
     *  Checks support of a given WGL extension, returns true if extension is supported.
     *
     *  @param[in] p_s_extension_name is null-terminated string, containing name
     *      of WGL extension encoded in UTF-8
     *
     *  @return Returns true in case p_s_extension_name is supported, otherwise false.
     */
    static bool b_SupportedWGLExtension(const char *p_s_extension_name);
#endif // _WIN32 || _WIN64

    /**
     *  @brief gets entry points of all extensions
     *
     *  Gets pointers of function entry points for all supported extensions and OpenGL versions.
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note Note this is aimed at simple and test implementations without
     *      much need for error checking. it is always better to call individual
     *      functions, because in case this fails, there's no telling
     *      which extension exactly failed to initialize.
     */
    static int n_GetAllSupportedExtensionsFuncPointers();

    /**
     *  @brief gets entry points of all OpenGL core functions
     *
     *  Gets pointers of function entry points for all supported OpenGL versions.
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note Note this is aimed at simple and test implementations without
     *      much need for error checking. it is always better to call individual
     *      functions, because in case this fails, there's no telling
     *      which extension exactly failed to initialize.
     */
    static int n_GetAllOpenGLCoreFuncPointers();

    /**
     *  @brief gets entry points of all OpenGL core functions, up to the version specified
     *
     *  Gets pointers of function entry points for all supported OpenGL versions, up to the version specified.
     *
     *  @param[in] n_major is required OpenGL version major
     *  @param[in] n_minor is required OpenGL version minor
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note Note this is aimed at simple and test implementations without
     *      much need for error checking. it is always better to call individual
     *      functions, because in case this fails, there's no telling
     *      which extension exactly failed to initialize.
     */
    static int n_GetOpenGLCoreFuncPointers(int n_major = 3, int n_minor = 3);

    /**
     *  @brief gets OpenGL 1.2 core function pointers, and sets GLEH_OPENGL_1_2
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL12FuncPointers();

    /**
     *  @brief gets OpenGL 1.3 core function pointers, and sets GLEH_OPENGL_1_3
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL13FuncPointers();

    /**
     *  @brief gets OpenGL 1.4 core function pointers, and sets GLEH_OPENGL_1_4
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL14FuncPointers();

    /**
     *  @brief gets OpenGL 1.5 core function pointers, and sets GLEH_OPENGL_1_5
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL15FuncPointers();

    /**
     *  @brief gets OpenGL 2.0 core function pointers, and sets GLEH_OPENGL_2_0
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL20FuncPointers();

    /**
     *  @brief gets OpenGL 2.1 core function pointers, and sets GLEH_OPENGL_2_1
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGL21FuncPointers();

    /**
     *  @brief gets OpenGL 3.0 core function pointers, and sets GLEH_OPENGL_3_0
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers for the following core extensions as well:
     *      GL_ARB_framebuffer_object, GL_ARB_map_buffer_range and GL_ARB_vertex_array_object.
     */
    static int n_GetGL30FuncPointers();

    /**
     *  @brief gets OpenGL 3.1 core function pointers, and sets GLEH_OPENGL_3_1
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers for the following core extensions as well:
     *      GL_ARB_uniform_buffer_object, GL_ARB_draw_instanced and GL_ARB_copy_buffer.
     */
    static int n_GetGL31FuncPointers();

    /**
     *  @brief gets OpenGL 3.2 core function pointers, and sets GLEH_OPENGL_3_2
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers for the following core extensions as well:
     *      GL_ARB_draw_elements_base_vertex, GL_ARB_provoking_vertex, GL_ARB_sync
     *      and GL_ARB_texture_multisample.
     */
    static int n_GetGL32FuncPointers();

    /**
     *  @brief gets OpenGL 3.3 core function pointers, and sets GLEH_OPENGL_3_3
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers from the following core extensions:
     *      GL_ARB_shader_bit_encoding, GL_ARB_blend_func_extended,
     *      GL_ARB_explicit_attrib_location, GL_ARB_occlusion_query2,
     *      GL_ARB_sampler_objects, GL_ARB_texture_rgb10_a2ui,
     *      GL_ARB_texture_swizzle, GL_ARB_timer_query, GL_ARB_instanced_arrays
     *      and GL_ARB_vertex_type_2_10_10_10_rev.
     */
    static int n_GetGL33FuncPointers();

    /**
     *  @brief gets OpenGL 4.0 core function pointers, and sets GLEH_OPENGL_4_0
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers from the following core extensions:
     *      GL_ARB_texture_query_lod, GL_ARB_draw_indirect, GL_ARB_gpu_shader5,
     *      GL_ARB_gpu_shader_fp64 GL_ARB_shader_subroutine,
     *      GL_ARB_tessellation_shader, GL_ARB_texture_buffer_object_rgb32,
     *      GL_ARB_texture_cube_map_array, GL_ARB_texture_gather
     *      GL_ARB_transform_feedback2 and GL_ARB_transform_feedback3.
     */
    static int n_GetGL40FuncPointers();

    /**
     *  @brief gets OpenGL 4.1 core function pointers, and sets GLEH_OPENGL_4_1
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note This gets function pointers from the following core extensions:
     *      ARB_ES2_compatibility, ARB_get_program_binary,
     *      ARB_separate_shader_objects, ARB_shader_precision,
     *      ARB_vertex_attrib_64bit and ARB_viewport_array.
     */
    static int n_GetGL41FuncPointers();